local log = Winning_Log

-------------------------------------------------------------------------------
-- WinningBgData
--
-- Encapsulates all Blizzard calls regarding BG data.
-------------------------------------------------------------------------------

WinningBgData = {}
WinningBgData.__index = WinningBgData

WinningBgData.new = function(self, player_name)
    local self = {}
    setmetatable(self, WinningBgData)

    self._player_name = player_name
    local found_player = false
    for idx = 1, GetNumBattlefieldScores() do
        local name, killingBlows, honorableKills, deaths, honorGained, faction, race, class, classToken, damageDone, healingDone, bgRating, ratingChange, preMatchMMR, mmrChange, talentSpec = GetBattlefieldScore(idx)
        if name == self._player_name then
            found_player = true
            self._kbs = killingBlows
            self._hks = honorableKills
            self._deaths = deaths
            self._damage = damageDone
            self._healing = healingDone
            self._class = classToken
            self._talent_spec = GetSpecializationInfo(GetSpecialization())
            self._race = race
            self._bg_faction = faction
            self._winning_faction = GetBattlefieldWinner() or 1
            self._bg_date = time()
            self._duration = math.floor(GetBattlefieldInstanceRunTime()/1000)

            SetMapToCurrentZone()
            self._map_id = GetCurrentMapAreaID()

            self._season = GetCurrentArenaSeason()
            self._is_rated = IsRatedBattleground()
        end
    end
    if not found_player then
        log("Player name is: " .. tostring(self._player_name))
        error("Could not find player in BG data: " .. self._player_name)
    end

    return self
end

WinningBgData.get_player_name = function(self)
    return self._player_name
end

WinningBgData.get_player_spec = function(self)
    return self._talent_spec
end

WinningBgData.get_player_race = function(self)
    return self._race
end

WinningBgData.get_date = function(self)
    return self._bg_date
end

WinningBgData.get_map_id = function(self)
    return self._map_id

end

WinningBgData.get_duration = function(self)
    return self._duration
end

WinningBgData.get_won = function(self)
    return self._bg_faction == self._winning_faction
end

WinningBgData.get_kbs = function(self)
    return self._kbs
end

WinningBgData.get_hks = function(self)
    return self._hks
end

WinningBgData.get_deaths = function(self)
    return self._deaths
end

WinningBgData.get_player_damage = function(self)
    return self._damage
end

WinningBgData.get_player_healing = function(self)
    return self._healing
end

WinningBgData.get_season = function(self)
    return self._season
end

WinningBgData.get_was_rated_match = function(self)
    return self._is_rated
end




-------------------------------------------------------------------------------
-- WinningWiper
--
-- Wipes data according to come context (ALL data, SEASON data, etc.)
-------------------------------------------------------------------------------

WinningWiper = {}

WinningWiper.init = function(self, addon)
    self.addon = addon
    self.db = self.addon.db.char.all_records
    self.bg_db = self.addon.db.char.bg_records
    self.stats = self.addon.db.char.stats
    self.ALL = "ALL"
    self.SEASON = "SEASON"
    self.PREVIOUS_SEASONS = "PREVIOUS_SEASONS"
    self._allowed_wipe_contexts = {
        [self.ALL] = self._wipe_all,
        [self.SEASON] = self._wipe_season,
        [self.PREVIOUS_SEASONS] = self._wipe_previous_seasons,
    }
    return self
end


--
-- Wipe bg record data according to a data context and a value.
--
WinningWiper.wipe = function(self, ctx, val)
    local wipe_ctx = self:_check_ctx(ctx)
    if wipe_ctx then
        -- This functions are not bound, have to call with self.
        return wipe_ctx(self, val)
    else
        -- what else to do here?
        return {}
    end
end


-- check wipe context
WinningWiper._check_ctx = function(self, ctx)
    local wipe_ctx = self._allowed_wipe_contexts[ctx]
    if wipe_ctx then
        log("Using wipe context: " .. ctx)
        return wipe_ctx
    else
        log(format("Unknown data context: %s", ctx))
    end
end


-- Wipe all data.
-- This is the only wipe method that actually wipes all BG data used in the UI,
-- for now - at least until I figure out the best way to handle the data.
WinningWiper._wipe_all = function(self, val)
    self:wipe_db()
    self:wipe_bg_db()
    self:wipe_stats()
end

WinningWiper.wipe_db = function(self)
    local counter = 0
    for i, v in pairs(self.db) do
        self.db[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d records.", counter))
end

WinningWiper.wipe_bg_db = function(self)
    local counter = 0
    for i, v in pairs(self.bg_db) do
        self.bg_db[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d bg records.", counter))
end

WinningWiper.wipe_stats = function(self)
    local counter = 0
    for i, v in pairs(self.stats) do
        self.stats[i] = nil
        counter = counter + 1
    end
    log(format("Wiped %d stat records.", counter))
end



--
-- WinningBattlegroundRecord model.
--
-- Just a helper to document what an expected bg model looks like.
WinningBattlegroundRecord = {
    PlayerName = false,
    PlayerSpec = false,
    PlayerRace = false,
    Type = false,
    Date = false,
    MapId = false,
    Map = false,
    DurationRaw = false,
    Duration = false,
    Won = false,
    Winner = false,
    KBs = false,
    HKs = false,
    Deaths = false,
    Damage = false,
    DamageRaw = false,
    Healing = false,
    HealingRaw = false,
    Season = false,
    Rated = false
}



-------------------------------------------------------------------------------
-- WinningDataAccessLayer
--
-- Control access to Record data.
-------------------------------------------------------------------------------

WinningDataAccessLayer = {}


WinningDataAccessLayer.init = function(self, addon)
    self.addon = addon

    -- This is the master record of data.
    self.db = addon.db.char.all_records

    -- This is a munged storage of data, fit for UI components.
    self.bg_db = addon.db.char.bg_records

    -- This is a continuously updated db of stats for ease of displaying.
    self.stats = addon.db.char.stats

    self.ALL = "ALL"
    self.SEASON = "SEASON"
    self.DAY = "DAY"
    self.MAP = "MAP"
    self._allowed_contexts = {
        [self.ALL] = self._filter_data_all,
        [self.SEASON] = self._filter_data_season,
        [self.DAY] = self._filter_data_day,
        [self.MAP] = self._filter_data_map,
    }

    return self
end


WinningDataAccessLayer.get_record = function(self, record_idx)
    return self.bg_db[record_idx]
end

-- save bg data
WinningDataAccessLayer.save = function(self, bg_record)
    for k, v in pairs(WinningBattlegroundRecord) do
        log("checking: " .. k)
        assert(bg_record[k] ~= nil)
    end
    log("Saving BG record into DB.")
    table.insert(self.db, bg_record)

    log("Stuffing BG record into BG DB.")
    self:update_bg_db(bg_record)

    log("Saving stats.")
    self:update_stats(bg_record)
end


WinningDataAccessLayer.convert_bg_record = function(self, bg_record)
    local ui_data = {}
    for i, v in pairs(self.addon.records_cols) do
        ui_data[i] = bg_record[v['name']]
    end
    return ui_data
end


-- Resync the data that is curated for the UI
WinningDataAccessLayer.resync = function(self, wiper)
    -- wipe the data stores for the UI
    wiper:wipe_bg_db()
    wiper:wipe_stats()

    for i, bg_record in pairs(self.db) do
        log("Resyncing record: " .. tostring(i))
        self:update_bg_db(bg_record)
        self:update_stats(bg_record)
    end
end


WinningDataAccessLayer.update_bg_db = function(self, bg_record)
    table.insert(self.bg_db, self:convert_bg_record(bg_record))
end


-- Update our stats.
WinningDataAccessLayer.update_stats = function(self, bg_record)
    self.stats.total_hks = (self.stats.total_hks or 0) + bg_record['HKs']
    if bg_record['HKs'] > (self.stats.top_hks or 0) then
        self.stats.top_hks = bg_record['HKs']
    end

    self.stats.total_kbs = (self.stats.total_kbs or 0) + bg_record['KBs']
    if bg_record['KBs'] > (self.stats.top_kbs or 0) then
        self.stats.top_kbs = bg_record['KBs']
    end

    self.stats.total_damage = (self.stats.total_damage or 0) + bg_record['DamageRaw']
    if bg_record['DamageRaw'] > (self.stats.top_damage or 0) then
        self.stats.top_damage = bg_record['DamageRaw']
    end

    self.stats.total_healing = (self.stats.total_healing or 0) + bg_record['HealingRaw']
    if bg_record['HealingRaw'] > (self.stats.top_healing or 0) then
        self.stats.top_healing = bg_record['HealingRaw']
    end
end


-- Max of a particular statistic for a given context.
WinningDataAccessLayer.max_stat_in_ctx = function(self, stat_key, ctx, val)
    local data = self:get_data(ctx, val)
    local val = 0
    for i, bg_record in pairs(data) do
        local new_val = bg_record[stat_key] or 0
        if new_val > val then
            val = new_val
        end
    end
    return val
end

-- Sum up all of a particular statistic for a given context.
WinningDataAccessLayer.summed_stat_in_ctx = function(self, stat_key, ctx, val)
    local data = self:get_data(ctx, val)
    local val = 0
    for i, bg_record in pairs(data) do
        local new_val = bg_record[stat_key] or 0
        val = val + new_val
    end
    return val
end


WinningDataAccessLayer.win_loss = function(self, ctx, val)
    local data = self:get_data(ctx, val)
    local wins = 0
    local losses = 0
    for i, bg_record in pairs(data) do
        if bg_record['Won'] then
            wins = wins + 1
        else
            losses = losses + 1
        end
    end
    return wins, losses
end


WinningDataAccessLayer.win_loss_ratio = function(self, wins, losses)
    if wins == 0 then
        return "0%"
    end
    if losses == 0 and wins > 0 then
        return "100%"
    end
    if wins > 0 and losses > 0 then
        return format("%d%%", math.floor((wins/(wins+losses))*100))
    end
end

-- Fetch data appropriate to the context.
WinningDataAccessLayer.get_data = function(self, ctx, val)
    local ctx_filter = self:check_ctx(ctx)
    if ctx_filter then
        -- This functions are not bound, have to call with self.
        return ctx_filter(self, val)
    else
        -- what else to do here?
        return {}
    end
end


WinningDataAccessLayer.check_ctx = function(self, ctx)
    local ctx_filter = self._allowed_contexts[ctx]
    if ctx_filter then
        log("Using data context: " .. ctx)
        return ctx_filter
    else
        log(format("Unknown data context: %s", ctx))
    end
end


WinningDataAccessLayer._filter_data_all = function(self, val)
    return self.db
end


WinningDataAccessLayer._filter_data_season = function(self, val)
    log(format("Filtering data for season: %s", val))
    local data = {}
    for i, r in pairs(self.db) do
        if r.Season == val then
            table.insert(data, r)
        end
    end
    return data
end


WinningDataAccessLayer._filter_data_day = function(self, val)
    local filter_date = date("%x", val)
    log(format("Filtering data for day: %s", filter_date))
    local data = {}
    for i, r in pairs(self.db) do
        if date("%x", r.Date) == filter_date then
            table.insert(data, r)
        end
    end
    return data
end


WinningDataAccessLayer._filter_data_map = function(self, val)
    log(format("Filtering data for map: %s", val))
    local data = {}
    for i, r in pairs(self.db) do
        if r.Map == val then
            table.insert(data, r)
        end
    end
    return data
end


